/*
 * Tcode9 virtual machine with X11 and audio extensions
 * Nils M Holm, 2017, 2018 CC0 license
 * https://creativecommons.org/publicdomain/zero/1.0/
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/select.h>
#include <sys/ioctl.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#ifdef ALSA
 #include <alsa/asoundlib.h>
#else
 #include <sys/soundcard.h>
#endif

#undef DEBUG

#define MINSIZE	131072
#define STKLEN	65536

#define byte	unsigned char
#define sbyte	signed char
#define cell	int

byte	*M;
sbyte	*S;
cell	Red;
cell	Memsize;

#define COLS	32
#define ROWS	24
#define WIDTH	(COLS*8)
#define HEIGHT	(ROWS*8)
#define NBMAPS	256		/* Don't change! */
#define MAXBMAP	16
#define MAXSCL	8

int	Scale = 0;
char	Keys[256];

#define DOWN	1
#define PRESSED	2

#define scale(x)	((x)*Scale)

Display		*Dsp;
int		Scr;
Window		Win;
GC		Gc;
Pixmap		Buf, Sbuf;
Cursor		Csr;
Colormap	Cmap;
XColor		Colors[16];
int		Fg, Bg;
int		Xfd;
Pixmap		Bmap[NBMAPS];
int		Bmw[NBMAPS], Bmh[NBMAPS];
int		Bp = 128;
byte		Map[ROWS+MAXBMAP/8][COLS+MAXBMAP/8];

#define DSPDEV	"/dev/dsp"
#define DSPFREQ	22050

byte	Dspbuf[DSPFREQ];
int	Fragsize;
int	Wave = 1;

int	Argc;
char	**Argv;
int	O_sound = 1;
int	O_mono;

byte	Cset[128][8] = {
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /*   */
	{ 0x00, 0x00, 0x00, 0x1f, 0x1f, 0x18, 0x18, 0x18 }, /* u/l corner */
	{ 0x00, 0x00, 0x00, 0xf8, 0xf8, 0x18, 0x18, 0x18 }, /* u/r corner */
	{ 0x18, 0x18, 0x18, 0xf8, 0xf8, 0x00, 0x00, 0x00 }, /* b/r corner */
	{ 0x18, 0x18, 0x18, 0x1f, 0x1f, 0x00, 0x00, 0x00 }, /* b/l corner */
	{ 0x00, 0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0x00 }, /* h bar */
	{ 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18 }, /* v bar */
	{ 0x00, 0x00, 0x00, 0xff, 0xff, 0x18, 0x18, 0x18 }, /* u connector */
	{ 0x18, 0x18, 0x18, 0xf8, 0xf8, 0x18, 0x18, 0x18 }, /* r connector */
	{ 0x18, 0x18, 0x18, 0xff, 0xff, 0x00, 0x00, 0x00 }, /* b connector */
	{ 0x18, 0x18, 0x18, 0x1f, 0x1f, 0x18, 0x18, 0x18 }, /* l connector */
	{ 0x18, 0x18, 0x18, 0xff, 0xff, 0x18, 0x18, 0x18 }, /* cross  */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /*   0% fill */
	{ 0x55, 0x00, 0x55, 0x00, 0x55, 0x00, 0x55, 0x00 }, /*  25% fill */
	{ 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa }, /*  50% fill */
	{ 0xaa, 0xff, 0xaa, 0xff, 0xaa, 0xff, 0xaa, 0xff }, /*  75% fill */
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff }, /* 100% fill */
	{ 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00 }, /* u fill */
	{ 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f }, /* r fill */
	{ 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff }, /* b fill */
	{ 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0 }, /* l fill */
	{ 0xf0, 0xf0, 0xf0, 0xf0, 0x00, 0x00, 0x00, 0x00 }, /* u/l fill */
	{ 0x0f, 0x0f, 0x0f, 0x0f, 0x00, 0x00, 0x00, 0x00 }, /* u/r fill */
	{ 0x00, 0x00, 0x00, 0x00, 0x0f, 0x0f, 0x0f, 0x0f }, /* b/r fill */
	{ 0x00, 0x00, 0x00, 0x00, 0xf0, 0xf0, 0xf0, 0xf0 }, /* b/l fill */
	{ 0x0f, 0x0f, 0x0f, 0x0f, 0xff, 0xff, 0xff, 0xff }, /* u/l inv fill */
	{ 0xf0, 0xf0, 0xf0, 0xf0, 0xff, 0xff, 0xff, 0xff }, /* u/r inv fill*/
	{ 0xff, 0xff, 0xff, 0xff, 0xf0, 0xf0, 0xf0, 0xf0 }, /* b/r inv fill */
	{ 0xff, 0xff, 0xff, 0xff, 0x0f, 0x0f, 0x0f, 0x0f }, /* b/l inv fill */
	{ 0xf0, 0xf0, 0xf0, 0xf0, 0x0f, 0x0f, 0x0f, 0x0f }, /* ul/br fill */
	{ 0x0f, 0x0f, 0x0f, 0x0f, 0xf0, 0xf0, 0xf0, 0xf0 }, /* ur/bl fill */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ???  */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /*   */
	{ 0x00, 0x18, 0x18, 0x18, 0x18, 0x00, 0x18, 0x00 }, /* ! */
	{ 0x00, 0x24, 0x24, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* " */
	{ 0x00, 0x24, 0x7e, 0x24, 0x24, 0x7e, 0x24, 0x00 }, /* # */
	{ 0x00, 0x08, 0x3e, 0x28, 0x3e, 0x0a, 0x3e, 0x08 }, /* $ */
	{ 0x00, 0x62, 0x64, 0x08, 0x10, 0x26, 0x46, 0x00 }, /* % */
	{ 0x00, 0x18, 0x24, 0x14, 0x38, 0x44, 0x7a, 0x00 }, /* & */
	{ 0x00, 0x08, 0x08, 0x10, 0x00, 0x00, 0x00, 0x00 }, /* ' */
	{ 0x00, 0x0c, 0x10, 0x10, 0x10, 0x10, 0x0c, 0x00 }, /* ( */
	{ 0x00, 0x30, 0x08, 0x08, 0x08, 0x08, 0x30, 0x00 }, /* ) */
	{ 0x00, 0x00, 0x24, 0x18, 0x7e, 0x18, 0x24, 0x00 }, /* * */
	{ 0x00, 0x00, 0x10, 0x10, 0x7c, 0x10, 0x10, 0x00 }, /* + */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x08, 0x10 }, /* , */
	{ 0x00, 0x00, 0x00, 0x00, 0x7e, 0x00, 0x00, 0x00 }, /* - */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00 }, /* . */
	{ 0x00, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x00 }, /* / */
	{ 0x00, 0x3c, 0x46, 0x4a, 0x52, 0x62, 0x3c, 0x00 }, /* 0 */
	{ 0x00, 0x08, 0x18, 0x28, 0x08, 0x08, 0x3e, 0x00 }, /* 1 */
	{ 0x00, 0x3c, 0x42, 0x02, 0x3c, 0x40, 0x7e, 0x00 }, /* 2 */
	{ 0x00, 0x3c, 0x42, 0x0c, 0x02, 0x42, 0x3c, 0x00 }, /* 3 */
	{ 0x00, 0x04, 0x0c, 0x14, 0x24, 0x7e, 0x04, 0x00 }, /* 4 */
	{ 0x00, 0x7e, 0x40, 0x7c, 0x02, 0x42, 0x3c, 0x00 }, /* 5 */
	{ 0x00, 0x3c, 0x40, 0x7c, 0x42, 0x42, 0x3c, 0x00 }, /* 6 */
	{ 0x00, 0x7e, 0x02, 0x04, 0x08, 0x10, 0x10, 0x00 }, /* 7 */
	{ 0x00, 0x3c, 0x42, 0x3c, 0x42, 0x42, 0x3c, 0x00 }, /* 8 */
	{ 0x00, 0x3c, 0x42, 0x42, 0x3e, 0x02, 0x3c, 0x00 }, /* 9 */
	{ 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x10, 0x00 }, /* : */
	{ 0x00, 0x00, 0x00, 0x10, 0x00, 0x10, 0x10, 0x20 }, /* ; */
	{ 0x00, 0x00, 0x08, 0x10, 0x20, 0x10, 0x08, 0x00 }, /* < */
	{ 0x00, 0x00, 0x00, 0x7e, 0x00, 0x7e, 0x00, 0x00 }, /* = */
	{ 0x00, 0x00, 0x10, 0x08, 0x04, 0x08, 0x10, 0x00 }, /* > */
	{ 0x00, 0x3c, 0x42, 0x04, 0x08, 0x00, 0x08, 0x00 }, /* ? */
	{ 0x00, 0x3c, 0x42, 0x5a, 0x5e, 0x40, 0x3c, 0x00 }, /* @ */
	{ 0x00, 0x3c, 0x42, 0x42, 0x7e, 0x42, 0x42, 0x00 }, /* A */
	{ 0x00, 0x7c, 0x42, 0x7c, 0x42, 0x42, 0x7c, 0x00 }, /* B */
	{ 0x00, 0x3c, 0x42, 0x40, 0x40, 0x42, 0x3c, 0x00 }, /* C */
	{ 0x00, 0x78, 0x44, 0x42, 0x42, 0x44, 0x78, 0x00 }, /* D */
	{ 0x00, 0x7e, 0x40, 0x7c, 0x40, 0x40, 0x7e, 0x00 }, /* E */
	{ 0x00, 0x7e, 0x40, 0x7c, 0x40, 0x40, 0x40, 0x00 }, /* F */
	{ 0x00, 0x3c, 0x42, 0x40, 0x5e, 0x42, 0x3c, 0x00 }, /* G */
	{ 0x00, 0x42, 0x42, 0x7e, 0x42, 0x42, 0x42, 0x00 }, /* H */
	{ 0x00, 0x3e, 0x08, 0x08, 0x08, 0x08, 0x3e, 0x00 }, /* I */
	{ 0x00, 0x02, 0x02, 0x02, 0x02, 0x42, 0x3c, 0x00 }, /* J */
	{ 0x00, 0x44, 0x48, 0x70, 0x48, 0x44, 0x42, 0x00 }, /* K */
	{ 0x00, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7e, 0x00 }, /* L */
	{ 0x00, 0x42, 0x66, 0x5a, 0x42, 0x42, 0x42, 0x00 }, /* M */
	{ 0x00, 0x42, 0x62, 0x52, 0x4a, 0x46, 0x42, 0x00 }, /* N */
	{ 0x00, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x3c, 0x00 }, /* O */
	{ 0x00, 0x7c, 0x42, 0x42, 0x7c, 0x40, 0x40, 0x00 }, /* P */
	{ 0x00, 0x3c, 0x42, 0x42, 0x4a, 0x46, 0x3c, 0x00 }, /* Q */
	{ 0x00, 0x7c, 0x42, 0x42, 0x7c, 0x44, 0x42, 0x00 }, /* R */
	{ 0x00, 0x3c, 0x40, 0x3c, 0x02, 0x42, 0x3c, 0x00 }, /* S */
	{ 0x00, 0x7f, 0x08, 0x08, 0x08, 0x08, 0x08, 0x00 }, /* T */
	{ 0x00, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3c, 0x00 }, /* U */
	{ 0x00, 0x42, 0x42, 0x42, 0x42, 0x24, 0x18, 0x00 }, /* V */
	{ 0x00, 0x42, 0x42, 0x42, 0x5a, 0x66, 0x42, 0x00 }, /* W */
	{ 0x00, 0x42, 0x24, 0x18, 0x18, 0x24, 0x42, 0x00 }, /* X */
	{ 0x00, 0x41, 0x22, 0x14, 0x08, 0x08, 0x08, 0x00 }, /* Y */
	{ 0x00, 0x7e, 0x04, 0x08, 0x10, 0x20, 0x7e, 0x00 }, /* Z */
	{ 0x00, 0x3c, 0x20, 0x20, 0x20, 0x20, 0x3c, 0x00 }, /* [ */
	{ 0x00, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x00 }, /* \ */
	{ 0x00, 0x3c, 0x04, 0x04, 0x04, 0x04, 0x3c, 0x00 }, /* ] */
	{ 0x00, 0x08, 0x14, 0x22, 0x00, 0x00, 0x00, 0x00 }, /* ^ */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x7f, 0x00 }, /* _ */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ??? */
	{ 0x00, 0x3c, 0x42, 0x42, 0x7e, 0x42, 0x42, 0x00 }, /* A */
	{ 0x00, 0x7c, 0x42, 0x7c, 0x42, 0x42, 0x7c, 0x00 }, /* B */
	{ 0x00, 0x3c, 0x42, 0x40, 0x40, 0x42, 0x3c, 0x00 }, /* C */
	{ 0x00, 0x78, 0x44, 0x42, 0x42, 0x44, 0x78, 0x00 }, /* D */
	{ 0x00, 0x7e, 0x40, 0x7c, 0x40, 0x40, 0x7e, 0x00 }, /* E */
	{ 0x00, 0x7e, 0x40, 0x7c, 0x40, 0x40, 0x40, 0x00 }, /* F */
	{ 0x00, 0x3c, 0x42, 0x40, 0x5e, 0x42, 0x3c, 0x00 }, /* G */
	{ 0x00, 0x42, 0x42, 0x7e, 0x42, 0x42, 0x42, 0x00 }, /* H */
	{ 0x00, 0x3e, 0x08, 0x08, 0x08, 0x08, 0x3e, 0x00 }, /* I */
	{ 0x00, 0x02, 0x02, 0x02, 0x02, 0x42, 0x3c, 0x00 }, /* J */
	{ 0x00, 0x44, 0x48, 0x70, 0x48, 0x44, 0x42, 0x00 }, /* K */
	{ 0x00, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7e, 0x00 }, /* L */
	{ 0x00, 0x42, 0x66, 0x5a, 0x42, 0x42, 0x42, 0x00 }, /* M */
	{ 0x00, 0x42, 0x62, 0x52, 0x4a, 0x46, 0x42, 0x00 }, /* N */
	{ 0x00, 0x3c, 0x42, 0x42, 0x42, 0x42, 0x3c, 0x00 }, /* O */
	{ 0x00, 0x7c, 0x42, 0x42, 0x7c, 0x40, 0x40, 0x00 }, /* P */
	{ 0x00, 0x3c, 0x42, 0x42, 0x4a, 0x46, 0x3c, 0x00 }, /* Q */
	{ 0x00, 0x7c, 0x42, 0x42, 0x7c, 0x44, 0x42, 0x00 }, /* R */
	{ 0x00, 0x3c, 0x40, 0x3c, 0x02, 0x42, 0x3c, 0x00 }, /* S */
	{ 0x00, 0x7f, 0x08, 0x08, 0x08, 0x08, 0x08, 0x00 }, /* T */
	{ 0x00, 0x42, 0x42, 0x42, 0x42, 0x42, 0x3c, 0x00 }, /* U */
	{ 0x00, 0x42, 0x42, 0x42, 0x42, 0x24, 0x18, 0x00 }, /* V */
	{ 0x00, 0x42, 0x42, 0x42, 0x5a, 0x66, 0x42, 0x00 }, /* W */
	{ 0x00, 0x42, 0x24, 0x18, 0x18, 0x24, 0x42, 0x00 }, /* X */
	{ 0x00, 0x41, 0x22, 0x14, 0x08, 0x08, 0x08, 0x00 }, /* Y */
	{ 0x00, 0x7e, 0x04, 0x08, 0x10, 0x20, 0x7e, 0x00 }, /* Z */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ??? */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ??? */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ??? */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ??? */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ??? */
};

void writes(char *s) {
	write(1, s, strlen(s));
}

void wlog(char *s) {
	write(2, s, strlen(s));
}

void fail(char *s) {
	wlog("xtcvm: ");
	wlog(s);
	wlog("\n");
	_exit(1);
}

#define BSIZE	1024

char	B[BSIZE];
int	Fd, C, K;

int rdch(void) {
	if (C >= K) {
		K = read(Fd, B, BSIZE);
		C = 0;
		if (K < 1) return -1;
	}
	C = C+1;
	return B[C-1] & 255;
}

cell rdwd(void) {
	cell	v;

	v = rdch();
	v = v | (rdch() << 8);
	v = v | (rdch() << 16);
	v = v | (rdch() << 24);
	return v;
}

void readblk(byte *v, int k) {
	int	i, c;

	for (i=0; i<k; i++) {
		c = rdch();
		if (-1 == c) fail("file too short");
		v[i] = c & 255;
	}
}

void load(char *s) {
	int	c;
	cell	tp, dp;

	Fd = open(s, O_RDONLY);
	if (Fd < 0) fail("could not open program");
	C = 0;
	K = 0;
	c = rdch();
	while (c != '\n' && c != -1)
		c = rdch();
	if (rdwd() != 0x39583354) fail("not a t3xvm program");
	tp = rdwd();
	dp = rdwd();
	Red = tp + dp;
	M = malloc(MINSIZE);
	S = (sbyte *) M;
	if (NULL == M) fail("not enough memory");
	readblk(M, tp+dp);
	if (rdch() != -1) fail("trailing garbage");
	close(Fd);
}

void mapscr(int x, int y, int w, int h, int v) {
	int	i, j;

	w /= scale(8);
	h /= scale(8);
	for (i = 0; i < w; i++)
		for (j = 0; j < h; j++)
			Map[y+j][x+i] = v;
}

void clrmap(void) {
	int	i;

	for (i=0; i<ROWS; i++)
		memset(Map[i], 0, COLS);
}

#ifdef ALSA

 #define PCM_DEVICE "default"

 snd_pcm_t	*Dspdev;

 void dspinit(void) {
	snd_pcm_hw_params_t	*hwp;
	int			p;

	if (snd_pcm_open(&Dspdev, PCM_DEVICE, SND_PCM_STREAM_PLAYBACK, 0) < 0)
		printf("cannot open pcm device\n");
	snd_pcm_hw_params_alloca(&hwp);
	snd_pcm_hw_params_any(Dspdev, hwp);
	if (snd_pcm_hw_params_set_access(Dspdev, hwp,
			SND_PCM_ACCESS_RW_INTERLEAVED) < 0)
		printf("cannot set interleaved mode\n");
	if (snd_pcm_hw_params_set_format(Dspdev, hwp,
			SND_PCM_FORMAT_S16_LE) < 0)
		printf("cannot set format\n");
	if (snd_pcm_hw_params_set_channels(Dspdev, hwp, 2) < 0)
		printf("cannot set channels number\n");
	p = DSPFREQ;
	if (snd_pcm_hw_params_set_rate_near(Dspdev, hwp, &p, 0) < 0)
		printf("can't set rate\n");
	if (snd_pcm_hw_params(Dspdev, hwp) < 0)
		printf("can't set harware parameters\n");
	snd_pcm_prepare(Dspdev);
 }

#else /* !ALSA */

 int	Dspdev;

 void dspinit(void) {
	char	msg[256];
	int	p;

	if (0 == O_sound) return;
	Dspdev = open(DSPDEV, O_WRONLY|O_FSYNC);
	if (Dspdev < 0) {
		sprintf(msg, "cannot open dsp device: %s", DSPDEV);
		fail(msg);
	}
	p = AFMT_S16_LE;
        if (ioctl(Dspdev, SNDCTL_DSP_SETFMT, &p) < 0)
		fail("cannot set audio format");
	p = 1;
        if (ioctl(Dspdev, SNDCTL_DSP_STEREO, &p) < 0)
		fail("cannot set stereo mode");
	p = DSPFREQ;
        if (ioctl(Dspdev, SNDCTL_DSP_SPEED, &p) < 0)
		fail("cannot set frequency");
 }

#endif /* !ALSA */

int bitmap(int bp, int w, int h, byte *data);

void xinit(void) {
	XTextProperty		name;
	XWindowAttributes	attr;
	char	*s;
	XEvent	e;
	XColor	c;
	int	i, x, y;
	char	*RGB[] = { "#000000", "#000080", "#008000", "#008080",
			   "#800000", "#800080", "#808000", "#a0a0a0",
			   "#202020", "#0000ff", "#00ff00", "#00ffff",
			   "#ff0000", "#ff00ff", "#ffff00", "#ffffff" };


	Dsp = XOpenDisplay(NULL);
	if (NULL == Dsp) fail("cannot open display");
	Scr = XDefaultScreen(Dsp);
	XGetWindowAttributes(Dsp, DefaultRootWindow(Dsp), &attr);
	if (0 == Scale) {
		for (i=1; i < 8; i++)
			if (	256*(i+1) > attr.width ||
				192*(i+1) > attr.height
			)
				break;
		Scale = i;
	}
	x = attr.width/2 - scale(256)/2;
	y = attr.height/2 - scale(192)/2;
        Win = XCreateSimpleWindow(Dsp, DefaultRootWindow(Dsp),
        			x, y, scale(WIDTH), scale(HEIGHT), 0,
        			BlackPixel(Dsp, Scr),
        			BlackPixel(Dsp, Scr));
	Buf = XCreatePixmap(Dsp, Win, scale(WIDTH), scale(HEIGHT),
				DefaultDepth(Dsp, Scr));
	Sbuf = XCreatePixmap(Dsp, Win, scale(WIDTH), scale(HEIGHT),
				DefaultDepth(Dsp, Scr));
        XSelectInput(Dsp, Win, KeyPressMask | KeyReleaseMask | ExposureMask);
	XMapWindow(Dsp, Win);
	do {
		XNextEvent(Dsp, &e);
	} while (e.type != Expose);
	Gc = XCreateGC(Dsp, Win, 0, NULL);
	if (NULL == Gc) fail("cannot create graphics context");
	Cmap = XCreateColormap(Dsp, Win, DefaultVisual(Dsp, Scr), AllocNone);
	for (i=0; i<16; i++)
		if (0 == XAllocNamedColor(Dsp, Cmap, RGB[i], &Colors[i], &c))
			fail("cannot allocate colors");
	XSetFillStyle(Dsp, Gc, FillSolid);
	XSetPlaneMask(Dsp, Gc, AllPlanes);
	XSetFunction(Dsp, Gc, GXcopy);
	XSetBackground(Dsp, Gc, Colors[0].pixel);
	XSetForeground(Dsp, Gc, Colors[0].pixel);
	XFillRectangle(Dsp, Buf, Gc, 0, 0, scale(WIDTH), scale(HEIGHT));
	XSetForeground(Dsp, Gc, Colors[15].pixel);
	Xfd = ConnectionNumber(Dsp);
	memset(Keys, 0, sizeof(Keys));
	clrmap();
	srand(time(NULL));
	for (i=0; i<128; i++)
		bitmap(i, 8, 8, Cset[i]);
	Csr = XCreatePixmapCursor(Dsp, Bmap[0], Bmap[0], &Colors[0],
				&Colors[0], 0, 0);
	XDefineCursor(Dsp, Win, Csr);
	s = "XTCVM";
	if (XStringListToTextProperty(&s, 1, &name) == 0)
		fail("cannot create window name");
	XSetWMName(Dsp, Win, &name);
	dspinit();
}

void xfini(void) {
	XCloseDisplay(Dsp);
}

int mapkey(int k) {
	/* XXX should use XGetKeyboardMapping() instead */
	k = XKeycodeToKeysym(Dsp, k, 0);
	switch (k) {
	case XK_Up:		return 1;
	case XK_Right:		return 2;
	case XK_Down:		return 3;
	case XK_Left:		return 4;
	case XK_BackSpace:	return 8;
	case XK_Tab:		return 9;
	case XK_Return:		return 13;
	case XK_Escape:		return 27;
	default:		return k & 255;
	}
}

void redraw(void) {
	XCopyArea(Dsp, Buf, Win, Gc, 0, 0, scale(WIDTH), scale(HEIGHT), 0, 0);
	XFlush(Dsp);
}

void handle_event(XEvent *e) {
	switch (e->type) {
	case KeyPress:
		Keys[mapkey(e->xkey.keycode)] |= (DOWN|PRESSED);
		break;
	case KeyRelease:
		Keys[mapkey(e->xkey.keycode)] &= ~DOWN;
		break;
	case Expose:
		redraw();
		break;
	}
}

int xkey(int k, int capture) {
	struct timeval	t;
	XEvent		e;
	fd_set		fds;
	int		p, ready;

	for (;;) {
		FD_ZERO(&fds);
		FD_SET(Xfd, &fds);
		t.tv_sec = t.tv_usec = 0;
		ready = select(Xfd+1, &fds, NULL, NULL, &t) != 0;
		if (ready) {
			XNextEvent(Dsp, &e);
			handle_event(&e);
		}
		else {
			if (capture) {
				p = (Keys[k & 255] & PRESSED) != 0;
				if (p) Keys[k & 255] &= ~PRESSED;
				return p;
			}
			else {
				return (Keys[k & 255] & DOWN) != 0;
			}
		}
	}
	return 0;
}

int xgetkey(void) {
	XEvent	e;
	int	k = 0;;

	for (;;) {
		XNextEvent(Dsp, &e);
		handle_event(&e);
		if (e.type == KeyPress)
			return mapkey(e.xkey.keycode);
	}
}

int bitmap(int bp, int w, int h, byte *data) {
	byte	m[MAXBMAP*MAXSCL][MAXBMAP*MAXSCL];
	byte	nbm[MAXBMAP*MAXBMAP];
	int	i, j, k, x, y, p, q;
	int	bpl;

	bpl = w / 8;
	y = 0;
	for (j = 0; j < h; j++) {
		x = 0;
		for (i = 0; i < bpl; i++) {
			for (k = 128; k > 0; k /= 2) {
				for (p = 0; p < scale(1); p++) {
					for (q = 0; q < scale(1); q++) {
						m[y+p][x+q] =
							(data[j*bpl+i]&k) != 0;
					}
				}
				x += scale(1);
			}
		}
		y += scale(1);
	}
	i = 0;
	k = 1;
	nbm[0] = 0;
	for (y = 0; y<scale(h); y++) {
		for (x = 0; x<scale(w); x++) {
			if (256 == k) {
				nbm[++i] = 0;
				k = 1;
			}
			nbm[i] |= (m[y][x]?k:0);
			k *= 2;
		}
	}
	Bmap[bp] = XCreateBitmapFromData(Dsp, Win, nbm, scale(w), scale(h));
	Bmw[bp] = w*Scale;
	Bmh[bp] = h*Scale;
	return bp;
}

int xbitmap(int w, int h, byte *data) {
	if (Bp >= NBMAPS)
		fail("too mant bitmaps");
	if (w > MAXBMAP || h > MAXBMAP)
		fail("bitmap too large");
	if (w % 8 != 0 || h % 8 != 0)
		fail("bitmap size not a multiple of 8");
	return bitmap(Bp++, w, h, data);
}

void setmono(void) {
	if (Fg < Bg) {
		XSetForeground(Dsp, Gc, Colors[0].pixel);
		XSetBackground(Dsp, Gc, Colors[O_mono].pixel);
	}
	else {
		XSetForeground(Dsp, Gc, Colors[O_mono].pixel);
		XSetBackground(Dsp, Gc, Colors[0].pixel);
	}
}

void xfg(int c) {
	Fg = c & 15;
	XSetForeground(Dsp, Gc, Colors[Fg].pixel);
	if (O_mono) setmono();
}

void xbg(int c) {
	Bg = c & 15;
	XSetBackground(Dsp, Gc, Colors[Bg].pixel);
	if (O_mono) setmono();
}

void xblit(int x, int y, int bm) {
	bm &= NBMAPS-1;
	mapscr(x, y, Bmw[bm], Bmh[bm], bm);
	x *= scale(8);
	y *= scale(8);
	XCopyPlane(Dsp, Bmap[bm], Buf, Gc, 0, 0, Bmw[bm], Bmh[bm], x, y, 1);
}

void xredraw(void) {
	redraw();
}

void xclear(int x, int y, int bm) {
	bm = bm & (NBMAPS-1);
	XSetForeground(Dsp, Gc, Colors[Bg].pixel);
	XFillRectangle(Dsp, Buf, Gc, scale(x*8), scale(y*8),
			Bmw[bm], Bmh[bm]);
	XSetForeground(Dsp, Gc, Colors[Fg].pixel);
	mapscr(x, y, Bmw[bm], Bmh[bm], 0);
}

void xclrscr(void) {
	XSetForeground(Dsp, Gc, Colors[Bg].pixel);
	XFillRectangle(Dsp, Buf, Gc, 0, 0, scale(WIDTH), scale(HEIGHT));
	XSetForeground(Dsp, Gc, Colors[Fg].pixel);
	clrmap();
}

int xscreen(int x, int y) {
	return Map[y][x];
}

void xdelay(int ms) {
	XFlush(Dsp);
	usleep(ms*1000);
}

int xrnd(int n) {
	return rand() % n;
}

int xprint(int x, int y, char *s) {
	int	i, k;

	k = strlen(s);
	for (i = 0; i < k; i++)
		xblit(x+i, y, toupper(s[i]));
}

int square_wave(int f, int x) {
	return x % f < f/2? 16384: -16384;
}

int triangle_wave(int f, int x) {
	x %= f*4;
	if (x < f) {
		return x * 32766 / f;
	}
	if (x < f * 3) {
		return -((x - f) * 32766 / f - 32767);
	}
	return ((x - f * 3) * 32766 / f) - 32767;
}

void xwave(int w) {
	Wave = w;
}

#ifdef ALSA

 void xbeep(int freq, int len) {
	int     i, v, k;

	if (0 == O_sound) {
		usleep(len*500);
		return;
	}
	snd_pcm_prepare(Dspdev);
	len = len * DSPFREQ / 1000;
	freq = DSPFREQ / freq;
	len = len / freq * freq;
	k = 0;
	for (i = 0; i < len; i++) {
		if (0 == Wave)
			v = square_wave(freq, i);
		else if (1 == Wave)
			v = triangle_wave(freq, i*4);
		else
			v = 0;
		Dspbuf[k++] =  v          & 0xff;
		Dspbuf[k++] = (v / 0x100) & 0xff;
		Dspbuf[k++] =  v          & 0xff;
		Dspbuf[k++] = (v / 0x100) & 0xff;
		if (k >= DSPFREQ) {
			snd_pcm_writei(Dspdev, Dspbuf, k/4);
			k = 0;
		}
	}
	if (k > 0) snd_pcm_writei(Dspdev, Dspbuf, k/4);
 }               

#else /* !ALSA */

 void xbeep(int freq, int len) {
	int	i, v, k;

	if (0 == O_sound) {
		usleep(len*500);
		return;
	}
	len = len * DSPFREQ / 1000;
	freq = DSPFREQ / freq;
	len = len / freq * freq;
	k = 0;
	for (i = 0; i < len; i++) {
		if (0 == Wave)
			v = square_wave(freq, i);
		else if (1 == Wave)
			v = triangle_wave(freq, i*4);
		else
			v = 0;
		Dspbuf[k++] =  v          & 0xff;
		Dspbuf[k++] = (v / 0x100) & 0xff;
		Dspbuf[k++] =  v          & 0xff;
		Dspbuf[k++] = (v / 0x100) & 0xff;
		if (k >= DSPFREQ) {
			write(Dspdev, Dspbuf, k);
			k = 0;
		}
	}
	if (k > 0) write(Dspdev, Dspbuf, k);
 }

#endif /* !ALSA */

void xscroll(int dir) {
	int	i;

	XSetForeground(Dsp, Gc, Colors[Bg].pixel);
	XFillRectangle(Dsp, Sbuf, Gc, 0, 0, scale(WIDTH), scale(HEIGHT));
	XSetForeground(Dsp, Gc, Colors[Fg].pixel);
	switch(dir) {
	case 1:	XCopyArea(Dsp, Buf, Sbuf, Gc, 0, scale(8) ,
				scale(WIDTH), scale(HEIGHT-8), 0, 0);
		for (i=0; i<ROWS-1; i++)
			memmove(Map[i], Map[i+1], COLS);
		memset(Map[ROWS-1], 0, COLS);
		break;
	case 2:	XCopyArea(Dsp, Buf, Sbuf, Gc, scale(8), 0,
				scale(WIDTH-8), scale(HEIGHT), 0, 0);
		for (i=0; i<ROWS; i++) {
			memmove(Map[i], &Map[i][1], COLS-1);
			Map[i][COLS-1] = 0;
		}
		break;
	case 3:	XCopyArea(Dsp, Buf, Sbuf, Gc, 0, 0 ,
				scale(WIDTH), scale(HEIGHT-8), 0, scale(8));
		for (i=ROWS-1; i>0; i--)
			memmove(Map[i], Map[i-1], COLS);
		memset(Map[0], 0, COLS);
		break;
	case 4:	XCopyArea(Dsp, Buf, Sbuf, Gc, 0, 0,
				scale(WIDTH-8), scale(HEIGHT), scale(8), 0);
		for (i=0; i<ROWS; i++) {
			memmove(&Map[i][1], Map[i], COLS-1);
			Map[i][0] = 0;
		}
		break;
	default: return;
		break;
	}
	XCopyArea(Dsp, Sbuf, Buf, Gc, 0, 0, scale(WIDTH), scale(HEIGHT), 0, 0);
}

void xpoint(int x, int y) {
	XFillRectangle(Dsp, Buf, Gc, scale(x), scale(y), scale(1), scale(1));
}

/*
 * Using Bresenham's algorithm because we need pixel scaling.
 * Shamelessly fetched from Rosettacode:
 * https://rosettacode.org/wiki/Bitmap/Bresenham%27s_line_algorithm#C
 */
void xline(int x0, int y0, int x1, int y1) {
	int	dx = abs(x1-x0);
 	int	sx = x0<x1? 1: -1;
	int	dy = abs(y1-y0);
	int	sy = y0<y1? 1: -1; 
  	int	e  = (dx>dy? dx: -dy)/2;
	int	e2;
 
	for(;;){
		xpoint(x0, y0);
		if (x0==x1 && y0==y1) break;
		e2 = e;
		if (e2 >-dx) { e -= dy; x0 += sx; }
		if (e2 < dy) { e += dx; y0 += sy; }
	}
}

void xbox(int x, int y, int dx, int dy) {
	int	t;

	if (dx < x) { t = x; x = dx; dx = t; }
	if (dy < y) { t = y; y = dy; dy = t; }
	XFillRectangle(Dsp, Buf, Gc, scale(x), scale(y), scale(dx-x)+1,
			scale(dy-y)+1);
}

cell	A, F, I, P;

cell w(cell a) {
	return  M[a+0]
	     | (M[a+1] << 8)
	     | (M[a+2] << 16)
	     | (M[a+3] << 24);
}

void Sw(cell a, cell w) {
	M[a+0] =  w        & 255;
	M[a+1] = (w >> 8)  & 255;
	M[a+2] = (w >> 16) & 255;
	M[a+3] = (w >> 24) & 255;
}

void push(cell x) {
	P -= 4;
	Sw(P, x);
}

cell pop(void) {
	P += 4;
	return w(P-4);
}

#define a()	w(I+1)
#define a2()	w(I+5)
#define s()	S[I+1]
#define s2()	w(I+2)

cell memscan(cell p, cell c, cell k) {
	cell	i;

	for (i=0; i<k; i++)
		if (M[p+i] == c)
			return i;
	return -1;
}

cell sys_call(cell n) {
	cell	r;

#ifdef DEBUG
	printf("SYSCALL(%d): %x %x %x %x\n", n, w(P+12), w(P+8), w(P+4), w(P));
#endif
	switch (n) {
	case  0: r = memcmp(&M[w(P+12)], &M[w(P+8)], w(P+4)); break;
	case  1: memcpy(&M[w(P+8)], &M[w(P+12)], w(P+4)); r = 0; break;
	case  2: memset(&M[w(P+12)], w(P+8), w(P+4)); r = 0; break;
	case  3: r = memscan(w(P+12), w(P+8), w(P+4)); break;
	case  4: r = creat((char *) &M[w(P+4)], 0644); break;
	case  5: r = open((char *) &M[w(P+8)], w(P+4)); break;
	case  6: r = close(w(P+4)); break;
	case  7: r = read(w(P+12), &M[w(P+8)], w(P+4)); break;
	case  8: r = write(w(P+12), &M[w(P+8)], w(P+4)); break;
	case  9: r = rename((char *) &M[w(P+8)], (char *)&M[w(P+4)]); break;
	case 10: r = remove((char *) &M[w(P+4)]); break;
	case 128:r = 0; xinit(); break;
	case 129:r = 0; xfini(); break;
	case 130:r = xkey(w(P+4), 0); break;
	case 131:r = xkey(w(P+4), 1); break;
	case 132:r = xgetkey(); break;
	case 133:r = xbitmap(w(P+12), w(P+8), &M[w(P+4)]); break;
	case 134:r = 0; xfg(w(P+4)); break;
	case 135:r = 0; xbg(w(P+4)); break;
	case 136:r = 0; xblit(w(P+12), w(P+8), w(P+4)); break;
	case 137:r = 0; xredraw(); break;
	case 138:r = 0; xclear(w(P+12), w(P+8), w(P+4)); break;
	case 139:r = 0; xclrscr(); break;
	case 140:r = xscreen(w(P+8), w(P+4)); break;
	case 141:r = 0; xdelay(w(P+4)); break;
	case 142:r = xrnd(w(P+4)); break;
	case 143:r = 0; xprint(w(P+12), w(P+8), &M[w(P+4)]); break;
	case 144:r = 0; xwave(w(P+4)); break;
	case 145:r = 0; xbeep(w(P+8), w(P+4)); break;
	case 146:r = 0; xscroll(w(P+4)); break;
	case 147:r = 0; xpoint(w(P+8), w(P+4)); break;
	case 148:r = 0; xline(w(P+16), w(P+12), w(P+8), w(P+4)); break;
	case 149:r = 0; xbox(w(P+16), w(P+12), w(P+8), w(P+4)); break;
	default: fail("bad system call"); r = 0; break;
	}
	return r;
}

void run(int k) {
	int	t;

	Memsize = Red + k + STKLEN;
	M = realloc(M, Memsize);
	S = (sbyte *) M;
	if (NULL == M) fail("not enough memory");
	P = Memsize;
	for (I = 0;; I++) {
#ifdef DEBUG
	printf("F=%08x P=%08x I=%08x %2x a=%08x A=%08x S0=%08x\n",
		F, P, I, M[I], a(), A, w(P));
#endif
	if (P < Red) fail("stack overflow");
	switch (M[I]) {
	case 0x80:
	case 0x00: push(A); break;
	case 0x81:
	case 0x01: A = 0; break;
	case 0x82: A = s(); I++; break;
	case 0x02: A = a(); I += 4; break;
	case 0x83:
	case 0x03: A = a(); I += 4; break;
	case 0x84: A = F+s(); I++; break;
	case 0x04: A = F+a(); I += 4; break;
	case 0x85:
	case 0x05: A = w(a()); I += 4; break;
	case 0x86: A = w(F+s()); I++; break;
	case 0x06: A = w(F+a()); I += 4; break;
	case 0x87:
	case 0x07: Sw(a(), A); I += 4; break;
	case 0x88: Sw(F+s(), A); I++; break;
	case 0x08: Sw(F+a(), A); I += 4; break;
	case 0x89:
	case 0x09: Sw(pop(), A); break;
	case 0x8a:
	case 0x0a: M[pop()] = (byte) A; break;
	case 0x8b:
	case 0x0b: t = a(); Sw(t, w(t)+a2()); I += 8; break;
	case 0x8c: t = s(); Sw(F+t, w(F+t)+s2()); I += 5; break;
	case 0x0c: t = a(); Sw(F+t, w(F+t)+a2()); I += 8; break;
	case 0x8d: P -= s(); I++; break;
	case 0x0d: P -= a(); I += 4; break;
	case 0x8e: P += s(); I++; break;
	case 0x0e: P += a(); I += 4; break;
	case 0x8f:
	case 0x0f: push(P); break;
	case 0x90:
	case 0x10: Sw(a(), P); I += 4; break;
	case 0x91:
	case 0x11: A = (A<<2) + pop(); break;
	case 0x92:
	case 0x12: A = w(A); break;
	case 0x93:
	case 0x13: A = A + pop(); break;
	case 0x94:
	case 0x14: A = M[A]; break;
	/* case 0x95: MARK */
	/* case 0x15: MARK */
	/* case 0x96: RESOLV */
	/* case 0x16: RESOLV */
	case 0x97:
	case 0x17: push(I+4); I += a(); I += 4; break;
	case 0x98:
	case 0x18: I += a(); I += 4; break;
	case 0x99:
	case 0x19: I += a(); I += 4; break;
	case 0x9a:
	case 0x1a: if (0 == A) { I += a(); } I += 4; break;
	case 0x9b:
	case 0x1b: if (0 != A) { I += a(); } I += 4; break;
	case 0x9c:
	case 0x1c: if (pop() >= A) { I += a(); } I += 4; break;
	case 0x9d:
	case 0x1d: if (pop() <= A) { I += a(); } I += 4; break;
	case 0x9e:
	case 0x1e: push(F); F = P; break;
	case 0x9f:
	case 0x1f: F = pop(); I = pop(); break;
	case 0xa0: exit(s()); break;
	case 0x20: exit(a()); break;
	case 0xa1:
	case 0x21: A = -A; break;
	case 0xa2:
	case 0x22: A = ~A; break;
	case 0xa3:
	case 0x23: A = 0==A? -1: 0; break;
	case 0xa4:
	case 0x24: A = pop() + A; break;
	case 0xa5:
	case 0x25: A = pop() - A; break;
	case 0xa6:
	case 0x26: A = pop() * A; break;
	case 0xa7:
	case 0x27: A = pop() / A; break;
	case 0xa8:
	case 0x28: A = pop() % A; break;
	case 0xa9:
	case 0x29: A = pop() & A; break;
	case 0xaa:
	case 0x2a: A = pop() | A; break;
	case 0xab:
	case 0x2b: A = pop() ^ A; break;
	case 0xac:
	case 0x2c: A = pop() << A; break;
	case 0xad:
	case 0x2d: A = (unsigned) pop() >> A; break;
	case 0xae:
	case 0x2e: A = pop() == A? -1: 0; break;
	case 0xaf:
	case 0x2f: A = pop() != A? -1: 0; break;
	case 0xb0:
	case 0x30: A = pop() < A? -1: 0; break;
	case 0xb1:
	case 0x31: A = pop() > A? -1: 0; break;
	case 0xb2:
	case 0x32: A = pop() <= A? -1: 0; break;
	case 0xb3:
	case 0x33: A = pop() >= A? -1: 0; break;
	/* case 0xb4: WORD */
	/* case 0x34: WORD */
	case 0xb5:
	case 0x35: A = sys_call(M[I+1]); I = pop(); break;
	default:   fail("invalid opcode"); break;
	}}
}

int size(void) {
	int	k, n;

	k = 0;
	for (I = 0;; I++) {
		n = 0;
		switch (M[I]) {
		case 0x8d: n = s(); I++; break;
		case 0x0d: n = a(); I += 4; break;
		case 0x90:
		case 0x10: I += 4; break;
		case 0x98:
		case 0x18: I += a(); I += 4; break;
		default:   return k;
		}
		k += n;
	}
}

void usage(void) {
	printf("\nUsage: xtcvm [-a 1|0] [-m 1-15] [-s 1-8] program\n\n");
	printf("-a  audio on/off [%d]\n", O_sound);
	printf("-m  monochrome mode [none]\n");
	printf("-s  pixel scaling [auto]\n\n");
	exit(1);
}

int arg(int j, int *opt) {
	if (j >= Argc-1) usage();
	*opt = atoi(Argv[j+1]);
	return j+1;
}

cell main(cell argc, char **argv) {
	int	i, j;

	Argc = argc;
	Argv = argv;
	for (j = 1; j < argc; j++) {
		if (argv[j][0] != '-') break;
		for (i = 1; argv[j][i]; i++) {
			switch (argv[j][i]) {
			case 'a':	j = arg(j, &O_sound);
					i = strlen(argv[j])-1;
					break;
			case 's':	j = arg(j, &Scale);
					i = strlen(argv[j])-1;
					if (Scale < 1 || Scale > MAXSCL)
						usage();
					break;
			case 'm':	j = arg(j, &O_mono);
					i = strlen(argv[j])-1;
					if (O_mono < 1 || O_mono > 15)
						usage();
					break;
			}
		}
	}
	if (j != argc-1) usage();
	load(argv[j]);
	run(size());
	return 0;
}
